using System;
using System.IO;
using System.Diagnostics;
using System.Collections.Generic;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using NLog;
using Perforce.P4;

namespace p4api.net.unit.test
{
    class ResetServerType
    {
        private bool unicodeReset = false;
        private bool nonUnicodeReset = false;
        public bool IsResetRequired(bool unicode)
        {
            if (unicode)
                return unicodeReset;
            return nonUnicodeReset;
        }
        public void SetResetRequired(bool unicode, bool value)
        {
            if (unicode)
                unicodeReset = value;
            else
                nonUnicodeReset = value;
        }

    }

    class Utilities
	{
        private static Logger logger = LogManager.GetCurrentClassLogger();
        private static ResetServerType ResetServer = new ResetServerType();
        private static Stopwatch stopWatchAllTests = null;
        private static Stopwatch stopWatch = null;
        private static int testCount = 0;

        public static void LogTestStart(TestContext testContext)
        {
            testCount++;
            logger.Info("====== TestName: {0}", testContext.TestName);
            stopWatch = Stopwatch.StartNew();
            if (stopWatchAllTests == null)
                stopWatchAllTests = Stopwatch.StartNew();
        }
        public static void LogTestFinish(TestContext testContext)
        {
            logger.Info("------ {0}: {1}, {2} ms", testContext.TestName, testContext.CurrentTestOutcome, stopWatch.ElapsedMilliseconds);
        }

        public static void LogAllTestsFinish()
        {
            logger.Info("@@@@@@ Time for tests: {0} - {1} s", testCount, stopWatchAllTests.Elapsed);
        }

        public static void ClobberDirectory( String path )
		{
			DirectoryInfo di = new DirectoryInfo( path );

			ClobberDirectory( di );
		}

		public static void ClobberDirectory(DirectoryInfo di)
		{
			string comSpec = Environment.GetEnvironmentVariable("ComSpec");
			Process Zapper = new Process();
			ProcessStartInfo si = new ProcessStartInfo(comSpec, "/c rd /S /Q " + di.FullName);
			si.WorkingDirectory = Path.GetDirectoryName(di.FullName);
			si.UseShellExecute = true;
			Zapper.StartInfo = si;
			try
			{
				Zapper.Start();
			}
			catch (Exception ex)
			{
				logger.Info("In ClobberDirectory, Zapper.Start() failed: {0}", ex.Message);
			}
			if (Zapper.HasExited == false)
			{
				Zapper.WaitForExit();
			}
			if (Directory.Exists(di.FullName))
			{
				bool worked = false;
				int retries = 0;
				do
				{
					if (!di.Exists)
						return;

					try
					{
						FileInfo[] files = di.GetFiles();

						foreach (FileInfo fi in files)
						{
							if (fi.IsReadOnly)
								fi.IsReadOnly = false;
							fi.Delete();
						}
						DirectoryInfo[] subDirs = di.GetDirectories();
						foreach (DirectoryInfo sdi in subDirs)
						{
							ClobberDirectory(sdi);
						}

						di.Delete();

						worked = true;
					}
					catch (Exception)
					{
						System.Threading.Thread.Sleep(100);
					}
					retries++;
				}
				while (!worked && retries < 2);
			}
		}

        static String p4d_cmd = "-p localhost:6666 -Id {1} -r {0} -L p4d.log -vserver=3";
        static String rsh_p4d_cmd = "rsh:p4d -r {0} -L p4d.log -vserver=3 -i";
        static String restore_cmd = "-r {0} -jr checkpoint.{1}";
        static String upgrade_cmd = "-r {0} -xu";
        static String generate_key_cmd = string.Empty;

		static string rubbishBin = "c:\\MyTestDir-rubbish-bin";

        public static Process DeploySSLP4TestServer(string path, bool Unicode)
        {
            System.Environment.SetEnvironmentVariable("P4SSLDIR", path);
            string test = System.Environment.GetEnvironmentVariable("P4SSLDIR");
            p4d_cmd = "-p ssl:6666 -Id UnitTestServer -r {0}";
            generate_key_cmd = "-Gc";
            return DeployP4TestServer(path, 1, Unicode);
        }

        public static Process DeployIPv6P4TestServer(string path, string tcp, bool Unicode)
        {
            p4d_cmd = "-p " + tcp + ":::1:6666 -Id UnitTestServer -r {0}";
            return DeployP4TestServer(path, 1, Unicode);
        }

		public static Process DeployP4TestServer(string path, bool Unicode, string testName="")
		{
			return DeployP4TestServer(path, 1, Unicode, testName);
		}

        public static string TestServerRoot(string baseRoot, bool Unicode)
        {
            String extra = "a";
            if (Unicode)
            {
                extra = "u";
            }
            var root = Path.Combine(baseRoot, extra, "server");
            logger.Debug("Server root: {0}", root);
            return root;
        }

        public static string TestRshServerPort(string baseRoot, bool Unicode)
        {
            String extra = "a";
            if (Unicode)
            {
                extra = "u";
            }
            var root = Path.Combine(baseRoot, extra, "server");
            root = String.Format(rsh_p4d_cmd, root);
            logger.Debug("Server root: {0}", root);
            return root;
        }

        public static string TestClientRoot(string baseRoot, bool Unicode)
        {
            String extra = "a";
            if (Unicode)
            {
                extra = "u";
            }
            var root = Path.Combine(baseRoot, extra, "clients");
            logger.Debug("Client root: {0}", root);
            return root;
        }

        public static void SetClientRoot(Repository rep, string TestDir, bool unicode, string ws_client)
        {
            Client c = rep.GetClient(ws_client, null);
            c.Root = Path.Combine(TestClientRoot(TestDir, unicode), ws_client);
            rep.UpdateClient(c);
            var syncOpts = new SyncFilesCmdOptions(SyncFilesCmdFlags.Force, 0);
            rep.Connection.Client.SyncFiles(syncOpts, new FileSpec(new DepotPath("//...")));
        }


        public static Process DeployP4TestServer(string path, int checkpointRev, bool Unicode, string testName = "")
		{
			String zippedFile = "a.exe";
			if (Unicode)
			{
				zippedFile = "u.exe";
			}
			return DeployP4TestServer(path, checkpointRev, zippedFile, Unicode, testName);
		}

		public static Process DeployP4TestServer(string path, int checkpointRev, string zippedFile, bool Unicode, string testName = "")
		{
            return DeployP4TestServer(path, checkpointRev, zippedFile, null, Unicode, testName);
		}

		public static Process DeployP4TestServer(string testRoot, int checkpointRev, string zippedFile, string P4DCmd, bool unicode, string testName = "")
        {
            logger.Info("DeployP4TestServer");
            if (!Directory.Exists(rubbishBin))
            {
                Directory.CreateDirectory(rubbishBin);
            }

            string assemblyFile = typeof(Utilities).Assembly.CodeBase;
            String unitTestDir = Path.GetDirectoryName(assemblyFile);

            String EnvPath = Environment.GetEnvironmentVariable("path");
            String CurWDir = Environment.CurrentDirectory;

            var zipBase = Path.GetFileNameWithoutExtension(zippedFile); // a or u
            var unzippedContentsDir = Path.Combine(rubbishBin, zipBase);
            var testServerRoot = Path.Combine(testRoot, zipBase, "server");
            var testClientsRoot = Path.Combine(testRoot, zipBase, "clients");
            CreateDir(testServerRoot, 10);
            try
            {
                Environment.CurrentDirectory = testServerRoot;
            }
            catch (Exception ex)
            {
                bool dirExists = Directory.Exists(testServerRoot);
                logger.Error("Can't cd to {0}, {1}", testServerRoot, ex.Message);
                return null;
            }

            int idx;
            if (unitTestDir.ToLower().StartsWith("file:\\"))
            {
                // cut off the file:\\
                idx = unitTestDir.IndexOf("\\") + 1;
                unitTestDir = unitTestDir.Substring(idx);
            }
            if ((idx = unitTestDir.IndexOf("TestResults")) > 0)
            {
                unitTestDir = unitTestDir.Substring(0, idx);
                if (unitTestDir.ToLower().Contains("bin\\debug") == false)
                {
                    unitTestDir = Path.Combine(unitTestDir, "bin\\debug");
                }
            }

            // Decide whether to reset server
            bool resetServerDir = ResetServer.IsResetRequired(unicode);
            ResetServer.SetResetRequired(unicode, false);

            ProcessStartInfo si;
            String msg;
            string unitTestZip = Path.Combine(unitTestDir, zippedFile);
            string targetTestZip = Path.Combine(unzippedContentsDir, zippedFile);
            if (!Directory.Exists(unzippedContentsDir))
            {
                // Copy file and unzip it
                CreateDir(unzippedContentsDir, 5);
                CopyFile(unitTestZip, targetTestZip);
                FileInfo fi = new FileInfo(targetTestZip);

                Process Unzipper = new Process();

                // unpack the zip
                si = new ProcessStartInfo(zippedFile);
                si.WorkingDirectory = unzippedContentsDir;
                si.Arguments = "-y";
                logger.Info("Unzipping {0} {1}", si.FileName, si.Arguments);

                Unzipper.StartInfo = si;
                Unzipper.Start();
                Unzipper.WaitForExit();

                // Copy unzipped directory tree
                CopyDirTree(unzippedContentsDir, testServerRoot);
            }
            else if (resetServerDir)
            {
                CopyDirTree(unzippedContentsDir, testServerRoot);
            }

            // Make sure no db.* files present
            logger.Info("Removing db.* from {0}", testServerRoot);
            try
            {
                foreach (string fname in Directory.GetFiles(testServerRoot, "db.*", SearchOption.TopDirectoryOnly))
                {
                    System.IO.File.Delete(fname);
                }
            }
            catch (Exception ex)
            {
                logger.Error(ex.Message);
                logger.Error(ex.StackTrace);
            }
            // Reset client directories
            DelDir(testClientsRoot);
            CreateDir(testClientsRoot, 5);
            foreach (string fname in Directory.GetDirectories(unzippedContentsDir, "*space*", SearchOption.TopDirectoryOnly))
            {
                var dirOnly = Path.GetFileName(fname);
                CopyDirTree(fname, Path.Combine(testClientsRoot, dirOnly));
            }

            if (p4d_cmd.Contains("ssl:"))
            {
                Process GenKeyandCert = new Process();

                // generate private key and certificate 
                si = new ProcessStartInfo("p4d");
                si.Arguments = generate_key_cmd;
                si.WorkingDirectory = testServerRoot;
                si.UseShellExecute = false;

                msg = si.Arguments;
                logger.Info(msg);

                GenKeyandCert.StartInfo = si;
                GenKeyandCert.Start();
                GenKeyandCert.WaitForExit();
            }

            // restore the checkpoint
            Process RestoreCheckPoint = new Process();
            si = new ProcessStartInfo("p4d");
            si.Arguments = String.Format(restore_cmd, testServerRoot, checkpointRev);
            si.WorkingDirectory = testServerRoot;
            si.UseShellExecute = false;
            logger.Info("{0} {1}", si.FileName, si.Arguments);

            RestoreCheckPoint.StartInfo = si;
            RestoreCheckPoint.Start();
            RestoreCheckPoint.WaitForExit();

            // upgrade the db tables
            Process UpgradeTables = new Process();
            si = new ProcessStartInfo("p4d");
            si.Arguments = String.Format(upgrade_cmd, testServerRoot);
            si.WorkingDirectory = testServerRoot;
            si.UseShellExecute = false;
            logger.Info("{0} {1}", si.FileName, si.Arguments);

            UpgradeTables.StartInfo = si;
            UpgradeTables.Start();
            UpgradeTables.WaitForExit();

            Process p4d = new Process();

            if (P4DCmd != null)
            {
                string P4DCmdSrc = Path.Combine(unitTestDir, P4DCmd);
                string P4DCmdTarget = Path.Combine(testServerRoot, P4DCmd);
                System.IO.File.Copy(P4DCmdSrc, P4DCmdTarget);

                // run the command to start p4d
                si = new ProcessStartInfo(P4DCmdTarget);
                si.Arguments = String.Format(testServerRoot);
                si.WorkingDirectory = testServerRoot;
                si.UseShellExecute = false;
                logger.Info("{0} {1}", si.FileName, si.Arguments);

                p4d.StartInfo = si;
                p4d.Start();
            } else
            {
                //start p4d
                si = new ProcessStartInfo("p4d");
                if (string.IsNullOrEmpty(testName))
                    testName = "UnitTestServer";
                si.Arguments = String.Format(p4d_cmd, testServerRoot, testName);
                si.WorkingDirectory = testServerRoot;
                si.UseShellExecute = false;
                logger.Info("{0} {1}", si.FileName, si.Arguments);

                p4d.StartInfo = si;
                p4d.Start();
            }
            Environment.CurrentDirectory = CurWDir;

            return p4d;
        }

        private static void CopyDirTree(string srcDir, string targDir)
        {
            logger.Info("Copying directory {0} to {1}", srcDir, targDir);
            // First Create all of the directories
            foreach (string dirPath in Directory.GetDirectories(srcDir, "*",
                SearchOption.AllDirectories))
                Directory.CreateDirectory(dirPath.Replace(srcDir, targDir));

            // Then copy all the files & Replaces any files with the same name
            foreach (string fname in Directory.GetFiles(srcDir, "*.*", SearchOption.AllDirectories))
            {
                var target = fname.Replace(srcDir, targDir);
                try
                {
                    System.IO.File.Copy(fname, target, true);
                }
                catch (Exception ex)
                {
                    logger.Debug(ex.Message);
                }
            }
            logger.Info("Finished copying {0} to {1}", srcDir, targDir);
        }

        private static void CopyFile(string srcFile, string targetFile)
        {
            int retries = 3;
            int delay = 1000; // initial delay 1 second
            while (retries > 0)
            {
                try
                {
                    System.IO.File.Copy(srcFile, targetFile);
                    break; //success
                } catch (Exception)
                {
                    System.Threading.Thread.Sleep(delay);
                    delay *= 2; // wait twice as long next time
                    retries--;
                }
            }
        }

        private static void CreateDir(string path, int retries)
        {
            while ((Directory.Exists(path) == false) && (retries > 0))
            {
                try
                {
                    Directory.CreateDirectory(path);
                    if (Directory.Exists(path))
                    {
                        break;
                    }
                    retries--;
                    System.Threading.Thread.Sleep(1000);
                } catch (Exception ex)
                {
                    retries--;
                    bool dirExists = Directory.Exists(path);
                    Trace.WriteLine(ex.Message);
                    if (dirExists)
                    {
                        break;
                    }
                    System.Threading.Thread.Sleep(200);
                }
            }
        }


        private static void DelDir(string dirPath)
        {
            if (!Directory.Exists(dirPath))
                return;
            try
            {
                Directory.Delete(dirPath, true);
            } catch
            {
                try
                {
                    // delete failed, try to rename it
                    Directory.Move(dirPath, string.Format("{0}-{1}", dirPath, DateTime.Now.Ticks));
                } catch
                {
                    // rename failed, try to clobber it (can be slow so last resort)
                    Utilities.ClobberDirectory(dirPath);
                }
            }
        }

        public static void RemoveTestServer(Process p, String testRoot, bool resetDepot=false, bool unicode=false)
        {
            logger.Info("RemoveTestServer");
            if (p != null)
            {
                if (!p.HasExited)
                    p.Kill();
                p.WaitForExit();
                // sleep for a second to let the system clean up
                System.Threading.Thread.Sleep(100);
            }
            MainTest.RememberToCleanup(rubbishBin);
            MainTest.RememberToCleanup(testRoot);
            // Flag for next unit test to reset depot - see DeployP4TestServer
            ResetServer.SetResetRequired(unicode, resetDepot);
        }

        private static void MoveDir(string srcDir, string targDir)
        {
            logger.Info("Moving {0} to {1}", srcDir, targDir);
            try
            {
                int retries = 60;
                while ((Directory.Exists(srcDir)) && (retries > 0))
                {
                    try
                    {
                        // Try to rename it
                        Directory.Move(srcDir, targDir);
                        //must have worked
                        break;
                    } catch
                    {
                        retries--;
                        System.Threading.Thread.Sleep(1000);
                        if (retries <= 0)
                        {
                            throw;
                        }
                    }
                }
            } catch (Exception ex)
            {
                logger.Info("In DeployP4TestServer, Directory.Move failed: {0}", ex.Message);
                // rename failed, try to clobber it (can be slow so last resort)
                Utilities.ClobberDirectory(srcDir);
            }
        }
    }
}
